unit WDosEpcs;

interface

{$IFDEF DESIGNPACKAGE}
uses
  SysUtils, Classes, Forms, WDosTimers;
{$ELSE}
uses
  WDosSysUtils, WDosClasses, WDosForms, WDosPorts, WDosTimers;
{$ENDIF}

const
  MaxSlaves =30;

  { Identifikationstyp }
  itEPC3351 =112;

type
  TStatusIdent =record
    IdLo, IdHi, StLo, StHi: Byte;
  end;

  { Kartenspeicher fr den Demo-Betrieb }
  TBytes =array[0..$7FF] of Byte;

  { Aufzhlungstypen fr die Kartenadresse }
  TEpcBaseAddress =(ep200H, ep210H, ep220H, ep230H,
                    ep240H, ep250H, ep260H, ep280H,
                    ep290H, ep2A0H, ep2E0H, ep300H,
                    ep330H, ep340H, ep350H, ep360H);
  { Epc status constants }
  TEpcStatus =(esUnknown, esInit, esSucoNet, esError);

  TSnClass =(snK1, snK2);
  TSnMode =(snMaster, snSlave);
  TSlaveAddress =2..MaxSlaves +1;
  TSlaves =1..MaxSlaves;

  TProtocolLength = type LongWord;

const
  { TwdxEpcCtrl }
  BaseAddressDefault =ep300H;
  WatchIntervalDefault =200;
  ModeDefault =snMaster;
  NetClassDefault =snK1;
  ProtocolDefault =$3FFFFFFF;
  SlaveAddressDefault =2;
  SlavesDefault =30;

type
  TwdxEpcCtrl = class (TComponent)
  private
    FActive: Boolean;
    FBaseAddress: TEpcBaseAddress;
    FBaseHexAddr: Word;
    FBytes: TBytes;
    FCancel: Boolean;
    FDemo: Boolean;
    FMode: TSnMode;
    FNetClass: TSnClass;
    FProtocol: TProtocolLength;
    FSlaveAddress: TSlaveAddress;
    FSlaves: TSlaves;
    FStatus: TEpcStatus;
    FStreamedActive: Boolean;
    FWatchDog: Boolean;
    FWatchInterval: Integer;
    procedure SetProtocol(const Value: TProtocolLength);
    function GetBusIdent(aSlave: TSlaveAddress): Word;
    function GetBusStatus(aSlave: TSlaveAddress): Word;
  protected
    procedure CheckLock;
    procedure Configure;
    function GetBusByt(aSlave: TSlaveAddress; aIndex: Byte): Byte;
    function GetDP(Index: Word): Byte;
    function GetDPR(Index: Word): Byte;
    function GetStatus: TEpcStatus;
    function GetStatusIdent(aSlave: TSlaveAddress): TStatusIdent;
    function IOAddr(aSlave, aIndex: Byte): Word;
    procedure Loaded; override;
    procedure PatternTest;
    procedure SetActive(Value: Boolean);
    procedure SetBaseAddress(Value: TEpcBaseAddress);
    procedure SetBusByt(aSlave: TSlaveAddress; aIndex: Byte; Value: Byte);
    procedure SetDemo(Value: Boolean);
    procedure SetDP(Index: Word; Value: Byte);
    procedure SetDPR(Index: Word; Value: Byte);
    procedure SetMode(Value: TSnMode);
    procedure SetNetClass(Value: TSnClass);
    procedure SetSlaveAddress(Value: TSlaveAddress);
    procedure SetSlaves(Value: TSlaves);
    procedure SetStatus(Value: TEpcStatus);
    procedure SetWatchDog(Value: Boolean);
    procedure SetWatchInterval(Value: Integer);
    procedure StartSuconet;
    procedure UpdateWatchDog;
    procedure WaitForConfig;
    procedure WaitForConfirm;
    procedure WaitForReset;
    procedure WaitForValue( aAddress: Word; aValue: Byte; aTime: LongInt);
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
    procedure Cancel;
    procedure Close; virtual;
    function IsRunning: Boolean;
    procedure Open; virtual;
    procedure Refresh;
    procedure Reset;
    property BusByt[aSlave: TSlaveAddress; aIndex: Byte]: Byte read GetBusByt write 
        SetBusByt;
    property DP[Index: Word]: Byte read GetDP write SetDP;
    property DPR[Index: Word]: Byte read GetDPR write SetDPR; default;

    property BusIdent[aSlave: TSlaveAddress]: Word read GetBusIdent;
    property BusStatus[aSlave: TSlaveAddress]: Word read GetBusStatus;

  published
    property Active: Boolean read FActive write SetActive;
    property BaseAddress: TEpcBaseAddress read FBaseAddress write SetBaseAddress;
    property Demo: Boolean read FDemo write SetDemo;
    property Mode: TSnMode read FMode write SetMode default ModeDefault;
    property NetClass: TSnClass read FNetClass write SetNetClass default NetClassDefault;
    property Protocol: TProtocolLength read FProtocol write SetProtocol;
    property SlaveAddress: TSlaveAddress read FSlaveAddress write SetSlaveAddress;
    property Slaves: TSlaves read FSlaves write SetSlaves;
    property Status: TEpcStatus read GetStatus write SetStatus;
    property WatchDog: Boolean read FWatchDog write SetWatchDog;
    property WatchInterval: Integer read FWatchInterval write SetWatchInterval default 
        WatchIntervalDefault;
  end;
  
  EEpcCtrlError = class (Exception)
  end;
  
implementation

{
************************************** TwdxEpcCtrl ***************************************
}
constructor TwdxEpcCtrl.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FWatchInterval :=WatchIntervalDefault;
  FMode :=ModeDefault;
  FProtocol :=ProtocolDefault;
  FSlaveAddress :=SlaveAddressDefault;
  FSlaves :=SlavesDefault;
  FNetClass :=NetClassDefault;
  BaseAddress :=BaseAddressDefault;
end;

destructor TwdxEpcCtrl.Destroy;
begin
  { EpcCtrl schlieen }
  Close;
  inherited Destroy;
end;

procedure TwdxEpcCtrl.Cancel;
begin
  FCancel :=True;
end;

procedure TwdxEpcCtrl.CheckLock;
begin
  if FActive then
    raise EEpcCtrlError.Create(
        'Operation bei aktivem Controller nicht ausfhrbar');
end;

procedure TwdxEpcCtrl.Close;
begin
  FActive :=False;
  {$IFNDEF DESIGNPACKAGE}
  Reset;
  {$ENDIF}
end;

procedure TwdxEpcCtrl.Configure;
begin
  { warten, bis Controller zum Konfigurieren bereit ist }
  WaitForConfig;
  { Konfiguration beginnen }
  DPR[511] :=0;
  { Identifikationswort (Lowbyte) setzen }
  DPR[0] :=itEPC3351;
  { Identifikationswort (Highbyte) abhngig
   von der SUCONET-Klasse setzen }
  if NetClass =snK1 then
    DPR[1] :=$80
  else
    DPR[1] :=$90;
  { in der Betriebsart Master die Anzahl der Slave-Stationen am Bus setzen }
  if Mode =snMaster then
    DPR[2] :=Slaves
  else
    DPR[2] :=0;
  { in der Betriebsart Master die Telegrammlnge aller Stationen einstellen }
  if (NetClass =snK1) and (Mode =snMaster) then
  begin
    { bit = 1 Lang-/ 0 Kurzprotokoll }
    with LongRec(Protocol) do
    begin
      DPR[3] :=System.Lo(Lo);
      DPR[4] :=System.Hi(Lo);
      DPR[5] :=System.Lo(Hi);
      DPR[6] :=System.Hi(Hi) and $3F;
    end;
  end;
  { Adresse dieser Station einstellen }
  if Mode =snMaster then
    DPR[511] :=1
  else
    DPR[511] :=SlaveAddress;
  { warten, bis Daten vom Controller akzeptiert wurden }
  WaitForConfirm;
end;

function TwdxEpcCtrl.GetStatusIdent(aSlave: TSlaveAddress): TStatusIdent;
var
  Base: Word;
begin
  with Result do
  begin
    Base :=(aSlave -2) shl 1;
    IdLo :=DPR[Base];
    IdHi :=DPR[Base +1];
    StLo :=DPR[Base +60];
    StHi :=DPR[Base +61];
  end;
end;

function TwdxEpcCtrl.IOAddr(aSlave, aIndex: Byte): Word;
begin
  Result :=aIndex +(aSlave -2) *13;
end;

function TwdxEpcCtrl.IsRunning: Boolean;
begin
  if Status =esSucoNet then
  begin
    if Demo then
      Result :=True
    else
    begin
      if (NetClass =snK2) or (Mode =snSlave) then
        Result :=(DPR[ 61] and $20 >0)
      else
        Result :=(DPR[510] and $01 >0);
    end
  end
  else
    Result :=False;
end;

procedure TwdxEpcCtrl.Loaded;
begin
  inherited Loaded;
  try
    Active :=FStreamedActive;
  except
    { let Delphi handle exceptions in design modus }
    if csDesigning in ComponentState then
      Application.HandleException(Self)
    else
      raise;
  end;
end;

procedure TwdxEpcCtrl.Open;
begin
  if not FActive then
  begin
    {$IFNDEF DESIGNPACKAGE}
    { 1. Karte zurcksetzen }
    Reset;
    { warten, bis Reset ausgefhrt wurde }
    WaitForReset;
    { Status der EPC-Karte berprfen }
    if Status <>esInit then
      raise EEpcCtrlError.Create(
          'Fehler beim Initialisieren: kein Reset-Status erreicht');
    { 2. Speichermuster prfen }
    PatternTest;
    { 3. Controller konfigurieren }
    if Demo then Status :=esSucoNet else Configure;
    { 4. SucoNet-Betriebssystem starten }
    StartSuconet;
    { 5. WatchDog aktivieren }
    UpdateWatchDog;
    {$ENDIF}
    { wenn erfolgreich }
    FActive :=True;
  end;
end;

procedure TwdxEpcCtrl.PatternTest;
var
  X: Integer;
  M: Word;
begin
  { Prfmuster auf Adressierungsfehler untersuchen }
  { Prfreihe initialisieren }
  X :=0;
  { Abbruch-Flag initialisieren }
  FCancel :=False;
  for M :=4 to 2046 do
  begin
    { andere Applikationen aktualisieren }
    if (M mod 100) =0 then
      Application.ProcessMessages;
    { Prfwert errechnen}
    X :=(11 *X +205) mod 251;
    { Adressierungsfehler prfen }
    if X <>DP[M] then
      raise EEpcCtrlError.Create('Prfmuster-Test: Adressierungsfehler');
    { Abbruch-Flag prfen }
    if FCancel then
      raise EEpcCtrlError.Create('Prfmuster-Test abgebrochen');
  end;
end;

procedure TwdxEpcCtrl.Refresh;
begin
  { WatchDog-Timer zurcksetzen }
  if Active and not Demo then
  begin
    {$IFNDEF DESIGNPACKAGE}
    Port[FBaseHexAddr] :=4;
    Port[FBaseHexAddr +$01] :=Succ(Port[FBaseHexAddr +$01]) and $FF;
    {$ENDIF}
  end;
end;

procedure TwdxEpcCtrl.Reset;
var
  M: Word;
  X: Integer;
begin
  { Controller auf Reset vorbereiten }
  DPR[511] :=255;
  if FDemo then
  begin
    for M :=$000 to $003 do
      DP[M] :=0;
    { Prffolge initialisieren und in Speicher eintragen }
    X :=0;
    for M :=$004 to $7FE do
    begin
      { Prfwert errechnen und zuweisen }
      X :=(11 *X +205) mod 251;
      DP[M] :=X;
    end;
    { Ausnahme: der Wert in $7FF wird nicht nach der Formel
     berechnet sondern von der EPC auf 0 gesetzt }
    DP[$7FF] :=0;
    { Status simulieren }
    Status :=esInit;
  end
  else
    {$IFNDEF DESIGNPACKAGE}
    Port[FBaseHexAddr +$02] :=0;
    {$ENDIF}
end;

procedure TwdxEpcCtrl.StartSuconet;
begin
  if NetClass =snK2 then
  begin
    DPR[253] :=DPR[253] and not $20;     { Run/Stop flag }
    DPR[250] :=60;                       { Segmentlnge }
  end
  else
    if Mode =snMaster then
      DPR[510] :=1
    else
      DPR[61] :=DPR[61] and not $20;
end;

procedure TwdxEpcCtrl.UpdateWatchDog;
begin
  if not FDemo then
  begin
    {$IFNDEF DESIGNPACKAGE}
    Port[FBaseHexAddr] :=3;
    if FWatchDog then Port[FBaseHexAddr +$01] :=WatchInterval div 100
                 else Port[FBaseHexAddr +$01] :=0;
    {$ENDIF}
  end;
end;

procedure TwdxEpcCtrl.WaitForConfig;
begin
  { Controller in die Betriebsart Konfiguration setzen }
  DPR[511] :=170;
  { warten, bis Controller fr die Konfiguration bereit ist }
  try
    WaitForValue( 511, 85, 1000);
  except
    on E: EEpcCtrlError do
    begin
      E.Message :='Konfigurationsfehler: ' +E.Message;
      raise;
    end;
  end;
end;

procedure TwdxEpcCtrl.WaitForConfirm;
begin
  { warten, bis Controller die Konfigurationsdaten akzeptiert hat }
  try
    WaitForValue( 511, 0, 1000);
  except
    on E: EEpcCtrlError do
    begin
      E.Message :='Fehler bei Datenbernahme: ' +E.Message;
      raise;
    end;
  end;
end;

procedure TwdxEpcCtrl.WaitForReset;
begin
  try
    WaitForValue( 511, 0, 3000);
  except
    on E: EEpcCtrlError do
    begin
      E.Message :='Fehler beim Rcksetzen: ' +E.Message;
      raise;
    end;
  end;
end;

procedure TwdxEpcCtrl.WaitForValue( aAddress: Word; aValue: Byte; aTime: LongInt);
var
  MaxTime: TDateTime;
begin
  { Abbruch-Flag initialisieren }
  FCancel :=False;
  { max. Prfzeit ermitteln }
  MaxTime :=Now +MsToDateTime(aTime);
  repeat
    Application.ProcessMessages;
    { Abbruch-Flag prfen }
    if FCancel then
      raise EEpcCtrlError.Create('Operation abgebrochen');
    { testen, ob Prfzeit abgelaufen ist }
    if Now >MaxTime then
      raise EEpcCtrlError.Create('Zeitberschreitung');
  until DPR[aAddress] =aValue;
  { Handshake-Signal aufgetreten }
end;

procedure TwdxEpcCtrl.SetActive(Value: Boolean);
begin
  if csLoading in ComponentState then
    FStreamedActive :=Value
  else
    if Value then Open else Close;
end;

procedure TwdxEpcCtrl.SetBaseAddress(Value: TEpcBaseAddress);
  
  const
    BaseAddresses: array[TEpcBaseAddress] of Word
      =($200, $210, $220, $230, $240, $250, $260, $280,
        $290, $2A0, $2E0, $300, $330, $340, $350, $360);
  
begin
  if Value <>FBaseAddress then
  begin
    CheckLock;
    FBaseAddress :=Value;
    FBaseHexAddr :=BaseAddresses[Value];
  end;
end;

function TwdxEpcCtrl.GetBusByt(aSlave: TSlaveAddress; aIndex: Byte): Byte;
begin
  Result :=DPR[126 +IOAddr(aSlave, aIndex)];
end;

procedure TwdxEpcCtrl.SetBusByt(aSlave: TSlaveAddress; aIndex: Byte; Value: Byte);
begin
  DPR[120 +IOAddr(ASlave, AIndex)] :=Value;
end;

procedure TwdxEpcCtrl.SetDemo(Value: Boolean);
begin
  if FDemo <> Value then
  begin
    CheckLock;
    FDemo := Value;
  end;
end;

function TwdxEpcCtrl.GetDP(Index: Word): Byte;
begin
  if FDemo then
    Result :=FBytes[Index]
  else
  begin
    {$IFDEF DESIGNPACKAGE}
    Result :=0;
    {$ELSE}
    Port[FBaseHexAddr] :=Lo(Index);
    Result :=Port[FBaseHexAddr + Succ(Hi(Index) shl 1)];
    {$ENDIF}
  end;
end;

procedure TwdxEpcCtrl.SetDP(Index: Word; Value: Byte);
begin
  if FDemo then
    FBytes[Index] :=Value
  else
  begin
    {$IFNDEF DESIGNPACKAGE}
    Port[FBaseHexAddr] :=Lo(Index);
    Port[FBaseHexAddr + Succ(Hi(Index) shl 1)] :=Value;
    {$ENDIF}
  end;
end;

function TwdxEpcCtrl.GetDPR(Index: Word): Byte;
begin
  Result :=DP[Index +$600];
end;

procedure TwdxEpcCtrl.SetDPR(Index: Word; Value: Byte);
begin
  DP[Index +$600] :=Value;
end;

procedure TwdxEpcCtrl.SetMode(Value: TSnMode);
begin
  if FMode <>Value then
  begin
    CheckLock;
    FMode :=Value;
  end;
end;

procedure TwdxEpcCtrl.SetNetClass(Value: TSnClass);
begin
  if FNetClass <>Value then
  begin
    CheckLock;
    FNetClass :=Value;
  end;
end;

procedure TwdxEpcCtrl.SetProtocol(const Value: TProtocolLength);
begin
  if FProtocol <>Value then
  begin
    CheckLock;
    FProtocol := Value and $3FFFFFFF;
  end;
end;

procedure TwdxEpcCtrl.SetSlaveAddress(Value: TSlaveAddress);
begin
  if FSlaveAddress <>Value then
  begin
    CheckLock;
    FSlaveAddress :=Value;
  end;
end;

procedure TwdxEpcCtrl.SetSlaves(Value: TSlaves);
begin
  if FSlaves <>Value then
  begin
    CheckLock;
    FSlaves :=Value;
  end;
end;

function TwdxEpcCtrl.GetStatus: TEpcStatus;
begin
  if FDemo then
    Result :=FStatus
  else
  begin
    {$IFDEF DESIGNPACKAGE}
    Result :=esUnknown;
    {$ELSE}
    Port[FBaseHexAddr]:=1;
    case Port[FBaseHexAddr+$01] of
    0:   Result :=esInit;
    2:   Result :=esSucoNet;
    255: Result :=esError;
    else
      Result :=esUnknown;
    end;
    {$ENDIF}
  end;
end;

procedure TwdxEpcCtrl.SetStatus(Value: TEpcStatus);
begin
  if Demo then
    FStatus :=Value;
end;

procedure TwdxEpcCtrl.SetWatchDog(Value: Boolean);
begin
  if FWatchDog <>Value then
  begin
    CheckLock;
    FWatchDog :=Value;
    UpdateWatchDog;
  end;
end;

procedure TwdxEpcCtrl.SetWatchInterval(Value: Integer);
begin
  if FWatchInterval <>Value then
  begin
    CheckLock;
    Value :=Value div 100;
    if Value <1 then Value :=1;
    if Value >255 then Value :=255;
    FWatchInterval :=Value *100;
    UpdateWatchDog;
  end;
end;

function TwdxEpcCtrl.GetBusIdent(aSlave: TSlaveAddress): Word;
var
  Addr: Word;
begin
  with WordRec(Result) do
  begin
    Addr :=(aSlave -2) *2;
    Lo :=DPR[Addr];
    Hi :=DPR[Succ(Addr)];
  end;
end;

function TwdxEpcCtrl.GetBusStatus(aSlave: TSlaveAddress): Word;
var
  Addr: Word;
begin
  with WordRec(Result) do
  begin
    Addr :=(aSlave -2) *2 +60;
    Lo :=DPR[Addr];
    Hi :=DPR[Succ(Addr)];
  end;
end;

end.
